#include "wizard.h"
#include "object.h"

#define REG "r1/obj/bull_board"
#define MAX_HEADER 30

/* Groo         90-12-02 */
/* Turvs/Arkon  91-07-11 */
/* Man-ing      92-02-22 */
/* Adapt to new mudlib, cleanup code. */
/* T/A -
 *  One board, settable as portable, wiz/player, boardname, message #s
 *  Added - copy (to personal notebook)
 *          move <#> <name> for moving to another board.
 *          permissions for each verb
 *          board status
 */

/*  Do not COPY!! clone or inherit as needed.  */

pointer messages;

string board_title,save_file;

int portable, read_level, write_level, erase_level, pax, lock, mess;

static string new_head, new_text;
static int line, looked_at;
static object curr_writer;
static object tow;

set_save_file(file) { save_file=file; }

set_pax(num) { pax = num; }

set_lock(num) { lock = num; }

set_mess(num) { mess = num; }

set_board_title(str) { board_title = str; }

set_erase_level(num) { erase_level = num; }

set_write_level(num) { write_level = num; }

set_portable(num) 
{  
   portable = num; 
   set_can_pick_up(1);
   set_can_drop(1);
   set_id("notebook:notepad:book:pad:note book:note pad");
   set_short("A Note Book");
   set_long("A Note Book.\n");
}

short()
{ 
  return short_desc+" Entitled: "+board_title;
}

long()
{
string tmp;
int i;

    tmp=long_desc;

    tmp+="Entitled: "+board_title+"\n";

    if(read_level) tmp+="["+read_level+"]   ";
    tmp+="Type 'read help' for usage info.\n";

    if(!sizeof(messages)) 
      {
        tmp+="The board is empty.\n";
        return tmp;
      }

    if(!read()) 
      {
        tmp+="The board dazzles you and you cannot read anything on it.\n";
        return tmp;
      }

    tmp+="The board contains "+sizeof(messages)+" note";

    if(sizeof(messages) > 1) tmp+="s";

    tmp+=" :\n\n";

    tmp+="Num  Title                          Author            Date & Time\n"; 
    tmp+="------------------------------------------------------------------------\n";

    for(;i<sizeof(messages);i++)
        sprintf(tmp,"%s%3d: %-30s %-11s %s\n",tmp,(i+1),messages[i][3],capitalize(messages[i][0]),messages[i][2]);
 
    return tmp;
}

help() 
{
int i;

    i = this_player()->query_level();
    write("Usage :\n");

    if(i>=read_level)
        write("        note <name>      To leave a note called <name>.\n");

    if(i>=write_level)
        write("        read <number>    To read note number <number>.\n");

    if(i>=erase_level)
        write("        remove <number>  To remove note number <number>.\n");

    if(pax || wiz())
        write("        copy <#>         To move <#> to your notebook.\n");

    if(wiz()) 
      {
        write("        store <#> <fn>   To store <#> in file <fn>.\n");
        write("        status board     To show board status.\n");
      }
    return 1;
}

init() 
{
int i;

    i = this_player()->query_level();

    if(i >= write_level)
        add_action("new_msg", "note");

    if(i >= read_level)
        add_action("read_msg", "read");

    if(i >= erase_level)
        add_action("remove_msg", "remove");

    if(pax || wiz())
        add_action("copy_msg", "copy");

    if(wiz())
      {
        add_action("store_msg", "store");
        add_action("status_this", "status");
      }

    if(!looked_at) restore_board();
}


new_msg(msg_head) 
{
    line = 1;

    if(!msg_head) return;

    if(!writ()) return 1;

    if(curr_writer && 
       (environment(curr_writer) == environment(this_object()) ||
       environment(curr_writer) == this_object() ||
       environment(this_object()) == curr_writer))
          write(curr_writer->query_name() + " is busy writing.\n");
             else if(sizeof(messages) == mess)
                write("You have to remove an old message first.\n");
                  else if(strlen(msg_head) > MAX_HEADER)
                    write("Message header too long. Try again.\n");
                      else
		       {
                         curr_writer = this_player();
                         say(curr_writer->query_name() + " starts writing a note.\n");
                         new_head = msg_head;
                         new_text = "";
                         input(this_player(),"get_msg");
                         write("Enter message text. End with '/s', abort with '/a'.\n");
                         write("1>>");
                       }
    return 1;
}

get_msg(str) 
{
    if(str == "/a") 
      {
        write("Note aborted.\n");
        clear();
        return;
      }

    if(str == "/s") 
      {
        if(line == 1) 
          {
            write("No text entered. Message discarded.\n");
            clear();
            return;
          }

        say(curr_writer->query_name()+" finishes a note : "+new_head+"\n");

        add( ({this_player()->query_real_name(), wiz(this_player()),
               ctime(time()), new_head, new_text}) );

        write("Ok.\n");
        clear();
        return;
      }

    new_text +=str+"\n";
    line++;
    write(line+">>");
    input(this_player(),"get_msg");
}

read_msg(what_msg) 
{
int i;

    if(what_msg == "help") 
      {
        help();
        return 1;
      }

    if(!read()) return 1;

    if(!sscanf(what_msg, "%d", i))
        if(!sscanf(what_msg, "%d", i))
            return;

    if(i<1 || i>sizeof(messages)) 
      {
        write("That message doesn't exist.\n");
        return 1;
      }

    i--;

    write("The note is titled '"+messages[i][3]+"':\n\n");
    write(messages[i][4]);
    return 1;
}

remove_msg(what_msg) 
{
    int i;

    if(!sscanf(what_msg, "%d", i))
        if(!sscanf(what_msg, "note %d", i))
            return;

    if(i<1 || i>sizeof(messages)) 
      {
       write("That message doesn't exits.\n");
       return 1;
      }

    i--;

    if(!erase(i)) return 1;

    if(lock_check(i)) return 1;

    say(this_player()->query_name()+" removes a note titled '"+
        messages[i][0]+"'.\n");

    rem(i);
    write("Ok.\n");
    return 1;
}

store_msg(str)
{
int i;
string s;

    if(!str) return;

    if(sscanf(str, "%d %s", i, s) != 2)
        if(sscanf(str, "note %d %s", i, s) != 2)
            if(sscanf(str, "%d", i) == 1)
                s = "/log/"+this_player()->query_real_name();
            else return;

    if(i<1 || i>sizeof(messages)) 
     {
        write("That message doesn't exist.\n");
        return 1;
     }

    i--;
    s += ".note";
    write_file(s, messages[i][3]+"\n"+messages[i][0]+"\t"+
                  messages[i][1]+"\t"+messages[i][2]+"\n"+messages[4]);

    write("Message ["+(i+1)+"] stored in "+s+"\n");
    return 1;
}

copy_msg(what_msg) 
{
int i;
string tmp;

    if(!what_msg || what_msg == "") return;

    tow = present("notebook", this_player());

    if(!tow)
      {
        write("You have no notebook.\n");
        return 1;
      }

    if(!sscanf(what_msg,"%s %d",tmp,i)) i=atoi(what_msg);

    if(tmp && tmp!="message") return;
     
    if(i<1 || i>sizeof(messages))  
      {
        write("That message doesn't exist.\n");
        return 1;
      }

    i--;

    if(lock_check(i) || !read(i) || !tow->writ(i)) 
       {
          write("You don't have access to copy this message.\n");
          return 1;
       }

    tow->add(messages[i]);
    write("Ok.\n");
    return 1;
}

add(txt) 
{
   if(!pointerp(messages)) messages=({txt});
    else messages+=({txt});

   save_board();
}

rem(num) 
{
    messages=del_element(messages,num);
    save_board();
}

save_board() 
{ 
  if(!save_file) return;
  save_object(save_file); 
}

lock_check(i) 
{
    if(lock) 
      {
        if((this_player()->query_real_name() != messages[i][0] &&
           wiz(this_player()) < messages[i][1]) ||
           !query_ip_number(this_player())) 
            {
              write("Only equal or higher level wizards may remove other players notes.\n");
              return 1;
            }
       }
}

clear() 
{
    curr_writer = 0;
    new_head = "";
    new_text = "";
}

writ() 
{
    if(this_player()->query_level() < write_level)
        write("You are not of a high enough level to write here.\n");
   else return 1;
}

read()
{
    if(this_player()->query_level() < read_level)
        write("You are not of a high enough level to read here.\n");
    else return 1;
}

erase() 
{
    if(this_player()->query_level() < erase_level)
       write("You are not of a high enough level to remove here.\n");
    else return 1;
}

status_this(str) 
{
    if(!id(str)) return;

    write("Level requirements -\n");
    write("Read : "+read_level);
    write("\tWrite : "+write_level);
    write("\tErase : "+erase_level+"\n");
    write("Board info -\n");
    write("Title : "+board_title+"\n");
    write(" "+sizeof(messages)+" messages, "+mess+" max.\n");

    if(pax || lock) write("This is a ");

    if(pax) 
      {
        write("public access");

        if(lock) write(", ");
      }

    if(lock) write("remove locked");

    if(pax || lock) write(" board.\n");

    if(portable) write("It is a notebook version.\n");

    return 1;
}

restore_board() 
{
    if(file_size("/"+save_file+".o")<1) return;
    restore_object(save_file);
    looked_at=1;
}

reset()
{
   set_id("board:bulletin board");
   set_short("A Bulletin Board");
   set_long("A Bulletin Board.\n");
   set_mess(20);
   return 1;
}
