#include <sys/types.h>
#include <sys/stat.h>
#ifdef M_XENIX
# include <sys/ndir.h>
#else
# ifdef AIX
#   include <dirent.h>
# else
#   include <sys/dir.h>
# endif
#endif
#ifdef M_XENIX
#include <signal.h>
#else
#include <sys/signal.h>
#endif
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include "lint.h"
#include "config.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "realms.h"
#include "sent.h"
#include "input_to.h"
#include "wiz_list.h"
#include "error.h"
#include "lan_fun.h"
#include "o_hash.h"
#include "s_hash.h"

#ifdef M_XENIX
void qsort();
#endif

extern REALMS *sys_realm;

extern char	*string_copy PROT((char *)), 
		    *make_shared_string PROT((char *));

extern int	yyparse(),
		stat(),
		fstat(),
		unlink(),
		link(),
		check_object(),
		_check_object(),
    		set_call PROT((OBJECT *, INPUT_TO *, int)),
		parse_command PROT((char *,OBJECT *)),
		print_call_out_usage();

extern void	ipc_remove(),
		clear_notify(),
		notify_no_command(),
		remove_ob_ref PROT((OBJECT *)),
		tell_npc PROT((OBJECT *,char *)),
		remove_all_players(),
		start_new_file PROT((FILE *)),
#ifdef MALLOC_smalloc
		dump_malloc_data(),
#endif
		print_lnode_status PROT((void));

extern OBJECT	*find_living_object PROT((char *,int)),
		*get_interactive_object PROT((int,int));

extern VALUE	*get_one_expr PROT((LNODE **,char *)),
		*print_expr PROT((LNODE *));

char		*check_file_name PROT((char *,int,int));

int		try_verb PROT((struct sentence *,char *)),legal_path(),
		special_parse(),match_string();

static void	check_for_recursion();

void		assign_vec(),
		add_message(),
		asign_svalue(),
		assign(),
		add_action(),
		print_local_commands PROT((OBJECT *,OBJECT *)),
		say(),
		tell_room(),
		shutdowngame(),
		free_svalue(),
		do_error(),
		shutdowngame();

extern OBJECT	*find_object PROT((char *)),
		*find_object2 PROT((char *,int));

VALUE		*users PROT((int)),*explode_string PROT((char *,char *));
OBJECT		*find_player PROT((char *,int));

extern struct object_status *ob_status[];

extern int 	errno,
		eval_cost,
		busy_calls;

int v_table_count=0,v_table_overhead=0;

char	shut_down_game=0;

struct sentence	*last_verb=0;

OBJECT		*command_giver;     /* Where the current command came from. */

int strchk(arg,null_to)
    register VALUE *arg;
    int null_to;
{
   if(!arg) return 1;
   if(arg->type!=T_STRING || !arg->u.string) return 1;
   if(null_to && !strcmp(arg->u.string,"\0")) return 1;
   return 0;
}

int numchk(arg,tp)
   register VALUE *arg;
   int tp;
{
  if(!arg) return 1;
  if(!(arg->type & T_NUMBER)) return 1;
  if(tp==T_NUMBER) return 0;
  if(arg->type!=(unsigned char)tp) return 1;
  return 0;
}

int objchk(arg)
  register VALUE *arg;
{
  if(!arg) return 1;
  if(arg->type!=T_OBJECT) return 1;
  if(!_check_object(arg->u.ob)) return 1;
  return 0;
}

int pointchk(arg)
  register VALUE *arg;
{
  if(!arg) return 1;
  if(arg->type!=T_POINTER) return 1;
  if(!arg->u.vec) return 1;
  return 0;
}

/*
 * Information about assignments of values:
 *
 * There are three types of l-values: Local variables, global variables
 * and vector elements.
 *
 * The local variables are stored in a vector, with pointers to values
 * of the 'struct value type'. Because they are local, they can all be
 * freed into the free list afterwards. When a local variable is used,
 * the pointer can be directly returned, as these values will never be
 * changed.
 *
 * The global variables must keep their values between executions, and
 * can't use the normal 'struct value', but will use a 'struct svalue' which
 * is smaller also. Because of this, no automatic garbage collection will
 * collect these values when they are overwritten with new values, and
 * thus their values will have to be freed immediately. When a global
 * variable is used, it must be copied into a new 'struct value'.
 *
 * Elements in vectors are eimilar to global variables. There is a reference
 * count to the whole vector, that states when to deallocate the vector.
 * The elements consists of 'struct svalue's, and will thus have to be freed
 * immediately when over written. When an item in vector is used,
 * it must be copied into a 'struct value'.
 */
 
VALUE *make_object(ob)
	register OBJECT *ob;
{
	register VALUE *ret;

	if(!check_object(ob)) return make_number(0,0.0,0);
	ret=alloc_value();
	ret->type=T_OBJECT;
	ret->u.ob=ob;
	ob->num_ref++;
	return ret;
}

VALUE *make_string(str)
    register char *str;
{
    register VALUE *ret = alloc_value();

    ret->type = T_STRING;
    ret->u.string = make_shared_string(str);
    return ret;
}

VALUE *make_number(n,n2,zero)
    int n;
    double n2;
    int zero;
{
    register VALUE *ret = alloc_value();

    if(n2 || (zero && !n)) 
	  {
		ret->type = T_FLOAT;
    	        ret->u.n_float = n2;
	  }
	else 
	  {
		ret->type=T_INT;
		ret->u.n_int=n;
	  }
    return ret;
}

VALUE *copy_svalue(arg)
    register SVALUE *arg;
{
    register VALUE *ret;

    if(!arg)
	  {
		 debug_message(sys_realm,"no value passed to copy_svalue()\n");
		 return &const0;
	   }
    switch(arg->type) {
    case T_STRING:
	ret=make_string(arg->u.string);
	break;
    case T_OBJECT:
	ret= make_object(arg->u.ob);
	break;
    case T_INT:
	ret=make_number(arg->u.n_int,0.0,0);
	break;
    case T_FLOAT: 
	ret=make_number(0,arg->u.n_float,1);
	break;
    case T_POINTER:
	ret = alloc_value();
	ret->type = T_POINTER;
	ret->u.vec = arg->u.vec;
	ret->u.vec->ref++;
	break;
	defualt:
		fatal("Invalid type to copy_svalue() %d\n",arg->type);
    }
    return ret;
}
/*
 * Allocate an array.
 */

int array_count=0;
int array_ovhead=0;

struct value *allocate_array(n)
    unsigned short n;
{
    register VALUE *v;
    int i;
    register VECTOR *p;
    if (n > cf.max_array_size)
	error(ob_status[0]->ob->realm,"Illegal array size.\n");
    i=sizeof (struct vector) + sizeof (struct value) * n-1;
    p = (struct vector *)xalloc(i);
    v = alloc_value();
    v->type = T_POINTER;
    v->u.vec = p;
    p->ref = 1;
    p->size =(unsigned short)n;
    p->sorted=0;
    p->real_size=(unsigned short)n;
    p->wl = ob_status[0]->ob->wl;
    p->wl->size_array += n;
    array_count+=n;
    array_ovhead+=i;
    for (i=0; i<n; i++) {
	p->item[i].type = T_INT;
	p->item[i].u.n_int = 0;
    }
    return v;
}

void free_vector(p)
    register VECTOR *p;
{
    int i;
	if(!p) return;
    p->ref--;
    if (p->ref > 0) return;
    for (i=0; i<p->size; i++) free_svalue(&p->item[i]);
    p->wl->size_array -= p->real_size;
    array_count-=p->real_size;
    array_ovhead-=sizeof (struct vector) + sizeof (struct svalue) * p->real_size-1;
    free((char *)p);
}

/*
 * Free the data that an svalue is pointing to. Not the svalue
 * itself.
 */
void free_svalue(v)
    register SVALUE *v;
{
    switch(v->type) {
    case T_STRING:
        if(!v->u.string)
	  debug_message(sys_realm,"no string in free_svalue\n");
	 else free_string(v->u.string,"free_svalue");
	break;
    case T_OBJECT:
	if(v->u.ob)
	   remove_ob_ref(v->u.ob);
	break;
    case T_POINTER:
	free_vector(v->u.vec);
    }
    v->type=T_INT;
    v->u.string=0;
    v->u.ob=0;
    v->u.vec=0;
    v->u.n_int=0;
    v->u.n_float=0.0;
}

/*
 * Assign a value to an svalue.
 * This is done either when element in vector, or when to an identifier
 * (as all identifiers are kept in a vector pointed to by the object).
 */

void _assign_svalue1(dest,v)
    register SVALUE *dest;
    register VALUE *v;
{
    dest->type = v->type;

    switch(dest->type)
     {
	case T_STRING:
		dest->u.string = make_shared_string(v->u.string);
		break;
	case T_OBJECT:
		dest->u.ob = v->u.ob;
		dest->u.ob->num_ref++;
		break;
	case T_POINTER:
		dest->u.vec = v->u.vec;
		dest->u.vec->ref++;
		break;
	case T_INT:
		dest->u.n_int = v->u.n_int;
		break;
	case T_FLOAT:
		dest->u.n_float=v->u.n_float;
		break;
	default:
		fatal("Bad data type '%d' to assign_svalue.\n",dest->type);
     }
}

void _assign_svalue2(dest,v)
    register SVALUE *dest;
    register SVALUE *v;
{
    
    dest->type = v->type;

    switch(dest->type)
     {
	case T_STRING:
		dest->u.string = make_shared_string(v->u.string);
		break;
	case T_OBJECT:
		dest->u.ob = v->u.ob;
		dest->u.ob->num_ref++;
		break;
	case T_POINTER:
		dest->u.vec = v->u.vec;
		dest->u.vec->ref++;
		break;
	case T_INT:
		dest->u.n_int = v->u.n_int;
		break;
	case T_FLOAT:
		dest->u.n_float=v->u.n_float;
		break;
	default:
		fatal("Bad data type '%d' to assign_svalue.\n",dest->type);
     }
}

void assign_svalue(dest, v2,v3)
    register SVALUE *dest;
    register VALUE *v2;
    register SVALUE *v3;
{
    /* First deallocate the previous value. */

    free_svalue(dest);

    if(v2) _assign_svalue1(dest,v2);
      else _assign_svalue2(dest,v3);
}

/*
 * Check that an assignment to an array item is not cyclic.
 */
static void check_for_recursion(vec, v)
    register VECTOR *vec, *v;
{
    register int i;
    if (vec->wl)
	vec->wl->cost++;
    eval_cost++;
    if (v == vec) error(ob_status[0]->ob->realm,"Recursive asignment of vectors.\n");
    for (i=0; i<v->size; i++) {
	if (v->item[i].type == T_POINTER)
	    check_for_recursion(vec, v->item[i].u.vec);
    }
}

/*
 * Assign a value to a vector element.
 */
void assign_vec(p, v)
    register LNODE *p;
    register VALUE *v;
{
    register VALUE *i, *vec;

    i = print_expr(p->a2);
    vec = print_expr(p->a1);
    if(!strchk(vec,0))
	  {
		char *temp;

		if(numchk(v,T_INT) || v->u.n_int<0 || v->u.n_int>255)
			 error(ob_status[0]->ob->realm,"Wrong type to assign to a string index.\n");

		if(numchk(i,T_INT) || i->u.n_int && (i->u.n_int<0 ||
		  i->u.n_int>=strlen(vec->u.string)))
			error(ob_status[0]->ob->realm,"Illegal index on string.\n");

		temp=string_copy(vec->u.string);
		temp[i->u.n_int]=(unsigned char)v->u.n_int;
		vec=make_string(temp);
		free(temp);

		assign(p->a1,vec);
		return;
	  }
    if (pointchk(vec)) error(ob_status[0]->ob->realm,"Indexing on illegal type. %d\n",vec->type);

    if (!pointchk(v)) check_for_recursion(vec->u.vec, v->u.vec);

    if (numchk(i,T_INT) || i->u.n_int<0 || i->u.n_int >= vec->u.vec->size)
	  error(ob_status[0]->ob->realm,"Illegal index on pointer.\n");

    assign_svalue(&vec->u.vec->item[i->u.n_int], v,0);
    vec->u.vec->sorted=0;
}
/*
 * Assign a value to a local variable or global variable.
 */

void assign(where, v)
    register LNODE *where;
    register VALUE *v;
{
    if (where->fun == f_argument) 
	fatal("No argument types should exist any longer.\n");

    if (where->fun == f_subscript) 
      {
   	     assign_vec(where, v);
	     return;
      }

    if (where->fun == f_local_name) 
      {
	struct lnode_number *p=(struct lnode_number*)where;

	if (!v) 
	     ob_status[0]->local_names[p->number] = &const0;
	   else ob_status[0]->local_names[p->number] = v;
	return;
      }

     if (where->fun == f_identifier) 
	{
  	 struct svalue *dest = find_value((struct lnode_number *)where);

	 assign_svalue(dest, v,0);
	 return;
        }

   error(ob_status[0]->ob->realm,"Assignment to illegal type.\n");
}

static VALUE *alloced_value_list=0;
static VALUE *value_free_list=0;

/*
 * Free all values previously allocated.
 */

void free_all_values() 
{
    while(alloced_value_list) 
      {
	register VALUE *tmp;

	tmp = alloced_value_list;
	alloced_value_list = alloced_value_list->next;

	if(tmp->type==T_STRING && tmp->u.string) 
	    if (!tmp->constant) free_string(tmp->u.string,"free_all_values");

	if(tmp->type==T_POINTER && tmp->u.vec) free_vector(tmp->u.vec);

	if(tmp->type==T_OBJECT && tmp->u.ob)
		 remove_ob_ref(tmp->u.ob);

	tmp->next=value_free_list;
	value_free_list=tmp;
     }
}


struct value *alloc_value() 
{
    register VALUE *tmp;

    if (!value_free_list) 
      {
	int i;

	value_free_list =(VALUE *)xalloc(cf.alloc_value * sizeof (VALUE));

	for (i=0; i<cf.alloc_value - 1; i++)
	    value_free_list[i].next  = &value_free_list[i+1];

	value_free_list[cf.alloc_value-1].next = 0;
    }

    tmp = value_free_list;
    value_free_list = value_free_list->next;

    /*
     * We preset the type to number, to ensure that no one forgets to
     * use the value, as it would have an old, undefined type when arriving
     * to free_all_values().
     */

#if 1
    (void)memset((char *)tmp,'\0',sizeof (VALUE));
#else
    tmp->u.string=0;
    tmp->u.n_int=0;
    tmp->u.n_float=0;
    tmp->u.vec=0;
    tmp->u.ob=0;
    tmp->constant=0;
#endif

    tmp->next = alloced_value_list;
    alloced_value_list = tmp;

    tmp->type = T_INT;

    return tmp;
}

struct svalue *find_value(p)
    register struct lnode_number *p;
{
#if 1  /* disable if you needed the debuging */

  if(!ob_status[0]->ip) return &ob_status[0]->ob->variables[p->number];

  return &ob_status[0]->ob->variables[ob_status[0]->ip->var_x_ref[p->number]];

#else  /* enable if you needed the debuging */

  if(!ob_status[0]) fatal("find_value: no current status\n");

  if(!ob_status[0]->ob) fatal("find_value: no current object\n");

  if(!ob_status[0]->ip)
    {

    if (p->number >= ob_status[0]->ob->num_variables) 
      {
	    fatal("Illegal variable access %d(%d) in ob:%s. See trace below.\n",
	         p->number, ob_status[0]->ob->num_variables,ob_status[0]->ob->name);
      }

    return &ob_status[0]->ob->variables[p->number];

    }
  else
    {
    if(!ob_status[0]->ob->inherit)
      {
	 debug_message(sys_relam,"Tried to x_ref a non inherited object\n");
	 debug_message(sys_realm,"co:%s io:%s\n",ob_status[0]->ob->name,ob_status[0]->ip->name);
	 ob_status[0]->ip=0;
	 return &ob_status[0]->ob->variables[p->number];
      }

    if(p->number>=ob_status[0]->ip->num_var)
      {
	  fatal("Illegal variable access %d(%d) in object:%s inherit prog:%s. See trace below.\n",
		p->number,ob_status[0]->ip->num_var,ob_status[0]->ob->name,ob_status[0]->ip->name);
      }

    return &ob_status[0]->ob->variables[ob_status[0]->ip->var_x_ref[p->number]];
    }

#endif
}

struct lnode_var_def *find_status(str)
    register char *str;
{
    register struct lnode_var_def *p;

    for (p=ob_status[0]->ob->status; p; p = p->next) {
        if(p->flags & L_STATIC || p->flags & L_PRIVATE) continue;
	if(!strcmp(str,p->name)) return p;
    }
    return 0;
}

extern struct lnode_2 *next_arg_list_to_use;

/*
 * Call function in another object. The object can be given both as a string
 * and a pointer.
 *
 * Save the caller object in ob_status[1]->ob so some functions can
 * see who called them.
 *
 * If find_object() is called, the object might be non-existing, and
 * thus loaded. If the object is loaded, "reset" will be called in the object.
 * next_arg_list_to_use must be hidden, so that the reset() won't
 * get these arguments.
 */
struct value *call_other(obj, name, arg)
    register VALUE *obj,*arg;
    register char *name;
{
    register OBJECT *ob, *save_command_giver = command_giver;
    register VALUE *ret;

    if (obj->type == T_OBJECT) {
	ob = obj->u.ob;
    } else {
	struct lnode_2 *save_args = next_arg_list_to_use;
	next_arg_list_to_use = 0;
	ob = find_object(obj->u.string);
	next_arg_list_to_use = save_args;
    }
    if (!ob)
	error(ob_status[0]->ob->realm,"Could not find object\n");

    if (ob->not_touched && ob->links[0]) {
	struct lnode_2 *save_args = next_arg_list_to_use;
	next_arg_list_to_use = 0;
	ob->not_touched=0;
	(void)apply(ob->reset, 0,ob, &const1,0);
	next_arg_list_to_use = save_args;
	command_giver = save_command_giver;
	if (!check_object(ob)) {
	    next_arg_list_to_use = 0;
	    return &const0;
	}
    }
    ret = apply(0,name, ob, arg,0);
    command_giver = save_command_giver;
    if (ret == 0)
	return &const0;
    return ret;
}

struct value *this_player() {
    register VALUE *p;

    if (!check_object(command_giver)) return &const0;
    p = make_object(command_giver);
    return p;
}

VALUE *this_object() {
    register VALUE *p;

    if (!check_object(ob_status[0]->ob)) return &const0;
    p = make_object(ob_status[0]->ob);
    return p;
}

char *make_new_name(str)
    register char *str;
{
    static int i;
    char *p = xalloc(strlen(str) + 10);

    (void)sprintf(p, "%s#%d", str, i);
    i++;
    return p;
}
/*
 * Execute a command for an object. Copy the command into a
 * new buffer, because 'parse_command()' can modify the command.
 * If the object is not current object(ob_status[0]->ob), static functions will
 * not be executed. This will prevent forcing players to do illegal things.
 */
int command_for_object(str, ob)
    register char *str;
    register OBJECT *ob;
{
    char buff[1000];

    if (strlen(str) > sizeof(buff) - 1)
	error(ob_status[0]->ob->realm,"Command too long.  Max: 1000.\n");

    if (!ob) ob = ob_status[0]->ob;

    (void)strcpy(buff, str);

    return parse_command(buff, ob);
}

/*
 * A message from an object will reach
 * all objects in the inventory,
 * all objects in the same environment and
 * the surrounding object.
 * Non interactive objects gets no messages.
 *
 * There are two cases to take care of. If this routine is called from
 * a player (indirectly), then the message goes to all in his
 * environment. Otherwise, the message goes to all in the current_object's
 * environment (as the case when called from a heart_beat()).
 *
 * If there is a second argument 'avoid_ob', then do not send the message
 * to that object.
 */

void say(v, avoid_ob,tr_flag)
    register VALUE *v;
    register OBJECT *avoid_ob;
    int tr_flag;
{
    register char *buff;
    register OBJECT *ob, *ob2, *save_command_giver = command_giver;
    register OBJECT *origin;
    register OBJECT *save_again;

    extern char *save_array PROT((VECTOR *,int));

    if (!tr_flag && ob_status[0]->ob->flags & COMMANDS) 
      command_giver = ob_status[0]->ob;

    if (command_giver) origin = command_giver;
      else origin = ob_status[0]->ob;

    if(v->type==T_STRING) buff=string_copy(v->u.string);
     else if(v->type==T_OBJECT) buff=string_copy(v->u.ob->name);
       else if(v->type==T_INT)
	 {
	   buff=xalloc(20);
	   (void)sprintf(buff, "%d", v->u.n_int);
	 }
        else if(v->type==T_FLOAT)
         {
	     buff=xalloc(200);
	     (void)sprintf(buff,"%lf",v->u.n_float);
	 }
        else if(v->type==T_POINTER) buff=save_array(v->u.vec,1);
            else do_error("1st","say"); 

    for (ob = origin->ref_ob[1]; ob; ob = ob->ref_ob[2]) 
       {

	if(!check_object(ob)) break;

	if (ob == command_giver || ob == avoid_ob ||
	    !(ob->flags & COMMANDS)) continue;

	if (!ob->interactive) tell_npc(ob, buff);

	save_again = command_giver;
	command_giver = ob;
	add_message("%s", buff);
	command_giver = save_again;
       }

  if(origin->ref_ob[0])
   {
    if (origin->ref_ob[0]->flags & COMMANDS)
      {
        if(origin != command_giver && origin != avoid_ob)
          {
	    if (origin->ref_ob[0]->interactive)
	      {
		save_again=command_giver;
	    	command_giver = origin->ref_ob[0];
	    	add_message("%s", buff);
		command_giver=save_again;
	      }
	    else tell_npc(origin->ref_ob[0], buff);
           }
       }

     for (ob = origin->ref_ob[0]->ref_ob[1]; ob; ob = ob->ref_ob[2])
        {
	  if(!check_object(ob)) break;

	  if (ob == avoid_ob || ob==command_giver || 
              !(ob->flags & COMMANDS)) continue;

	  if (!ob->interactive) tell_npc(ob, buff);
            else 
             {
	    	save_again = command_giver;
	   	command_giver = ob;
	    	add_message("%s", buff);
	    	command_giver = save_again;
	     }
         }
    }

    command_giver = save_command_giver;
    free(buff);
}

/*
 * Send a message to all objects inside an object.
 * Non interactive objects gets no messages.
 * See say().
 */

void tell_room(room, v ,avoid_ob)
    register OBJECT *room, *avoid_ob;
    register VALUE *v;
{
    register OBJECT *save_command_giver=command_giver;
    command_giver=room;
    say(v,avoid_ob,1);
    command_giver=save_command_giver;
}

void shout_string(str,sys_flag)
    register char *str;
    int sys_flag;
{
    register struct o_table_entry *tmp;
    register OBJECT *ob, *save_command_giver = command_giver;
    int i;
 
    extern int num_player;
    
    if(ob_status[0]->ob && 
	   ob_status[0]->ob->realm==sys_realm) sys_flag=1;
	if(ob_status[0] && command_giver!=ob_status[0]->ob)
	  {
        if(ob_status[0]->ob && ob_status[0]->ob->flags & COMMANDS) 
	       command_giver=ob_status[0]->ob;
	    else command_giver=0;
	  }

 if(!sys_flag)
   for(i=0;i<cf.otable;i++)
     for(tmp=o_table[i];tmp;tmp=tmp->next)  

	{  ob=tmp->ob;

           if (!(ob->flags & COMMANDS) || ob == command_giver ||
	       ob->flags & BLOCK_SHOUTS || ob->interactive ||
	       ob==ob_status[0]->ob || ob_status[0]->ob->realm!=ob->realm)
       		continue;
             else tell_npc(ob,str);
        }
    for(i=1;i<=num_player;i++)
     {
	   OBJECT *sh_ob;
	   OBJECT *p_save_command_giver=command_giver;

	   sh_ob=get_interactive_object(i,0);
       if(command_giver==sh_ob || 
	      command_giver==get_interactive_object(i,1) ||
	      sh_ob->flags & BLOCK_SHOUTS || (!sys_flag && ob_status[0]->ob && 
		  ob_status[0]->ob->realm!=sh_ob->realm))
               continue;
       command_giver = sh_ob;
       add_message("%s", str);
	   command_giver=p_save_command_giver;
     }
    command_giver = save_command_giver;
}

/*
 * Set up a function in this object to be called with the next
 * user input string.
 */
struct value *input_to(ob,fun, flag, arg)
    register OBJECT *ob;
    register char *fun;
    int flag;
    VALUE *arg;
{
    register INPUT_TO *s=alloc_input_to();
    struct fun_p *fun_p;

    fun_p=find_fun_p(fun,ob_status[0]->ob,0);
    if(!fun_p)
       {
	    free_input_to(s);
	    error(ob_status[0]->ob->realm,"input() to bad function.\n");
       }

    if(fun && set_call(ob, s, flag)) 
	  {
		s->ob = ob_status[0]->ob;
		s->fun_p=fun_p;
		s->ob->num_ref++;
	        if(arg)
                 {
		    s->arg.type=arg->type;
		    switch(arg->type)
	             {
		       case T_OBJECT:
		        s->arg.u.ob=arg->u.ob;
		        arg->u.ob->num_ref++;
			break;
		       case T_INT:
		        s->arg.u.n_int=arg->u.n_int;
			break;
		       case T_FLOAT:
		        s->arg.u.n_float=arg->u.n_float;
			break;
		       case T_STRING:
		        s->arg.u.string=make_shared_string(arg->u.string);
			break;
		       case T_POINTER:
		        s->arg.u.vec=arg->u.vec;
		        arg->u.vec->ref++;
		     }
		 }
		return &const1;
      }
    free_input_to(s);
    return &const0;
}

int pstrcmp(p1, p2)
    register char **p1, **p2;
{
    return strcmp(*p1, *p2);
}

/*
 * List files in a directory. The standard 'ls' could be used, but
 * it takes too much time.
 * Prepared for flag decoding.
 *
 * Look at the the last component of the path name. It is a regular
 * expression, select only matching files.
 *
 * All .i files and '.' and '..' are only listed if they match the regexp.
 */

void list_files(path)
    register char *path;
{
    char buff[160];	/* Will never need more than two lines */
    int i,num,j,pos,f_size;
    unsigned short max;
    register char **names,*p,*name_buffer;
    register DIR *dirp;
#ifdef AIX
    register struct dirent *de;
#else
    register struct direct *de;
#endif
    struct stat st;
    char regexp[80], path2[255],stat_file[255];
    int do_match = 0;

    if (!path) path = ".";

    path = check_file_name(path, 0,0);

    if (!path)
      {
	 add_message("No such dir, or lack of permisions.\n");
         return;
      }

    (void)strncpy(path2, path, sizeof path2);

    path2[sizeof path2 - 1] = '\0';

    p = path2 + strlen(path2) - 2;

    if (p >= path2 && p[0] == '/' && p[1] == '.') p[0] = '\0';

    if (stat(path2, &st) == -1) {
	/* Either the directory does not exist, or it is a regexp
	 * file name. Strip of the last component.
	 */

	p = strrchr(path2, '/');

	if (!p || p[1] == '\0') return;

	(void)strncpy(regexp, p+1, sizeof regexp);
	regexp[sizeof regexp - 1] = '\0';

	*p = '\0';
	do_match = 1;
    }

    dirp = opendir(path2);
    if(!dirp && errno == EMFILE) fatal("Max files open.\n");

    if (!dirp) {
	add_message("No such directory '%s'\n", path2);
	return;
    }

    max = 0;
    num = 0;

    /*
     * First step, find the number of file names.
     */

    for(de = readdir(dirp); de; de = readdir(dirp)) {

	if (!do_match && de->d_name[de->d_namlen-1] == 'i' &&
	    de->d_name[de->d_namlen-2] == '.')
	    continue;

	if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, "..")) continue;

	if (do_match && !match_string(regexp, de->d_name))
	    continue;

	if (de->d_namlen > max) max = de->d_namlen;

	num++;
    }

    if (!num)
      {
	closedir(dirp);
	return;
      }

    max+=3;

    /*
     * Copy all file names into a name space, and set up pointer
     * into it.
     */

    names = (char **)xalloc(num * sizeof (char *));
    name_buffer=xalloc(num * max+1);
    rewinddir(dirp);

    for(pos=0, i=0, de = readdir(dirp); i<num ;de = readdir(dirp)) {

	if (!do_match && de->d_name[de->d_namlen-1] == 'i' &&
	    de->d_name[de->d_namlen-2] == '.')
	    continue;

	if (!strcmp(de->d_name, ".") || !strcmp(de->d_name, "..")) continue;

	if (do_match && !match_string(regexp, de->d_name))
	    continue;

	names[i] = name_buffer + pos;

	(void)strcpy(name_buffer + pos, de->d_name);

	pos += de->d_namlen;

	(void)strcpy(stat_file,path2);
	(void)strcat(stat_file,"/");
	(void)strcat(stat_file,de->d_name);

	(void)stat(stat_file,&st);

	if(S_IFDIR & st.st_mode)
	  {
	     (void)strcpy(name_buffer+pos,"/");
	     pos+=2;
	  }
	else pos++;

      i++;
    }

    closedir(dirp);

    /* Sort the names. */

    qsort((char *)names, num, sizeof (char *), pstrcmp);

    /*
     * Print the file names in nice columns, as many as possible
     * on every line. That means that there can be one extra name
     * on each line, depending on the size of the last name on the line.
     */

    for(max+=6,i=0, j=0; j<num; j++) {
	char stat_file[255];

	if (i * max + strlen(names[j])+6 >= 79) {

	    /*
	     * Strip the last spaces on the last name to ensure that
	     * it will fit last on the line.
	     */

	    for (i=strlen(buff) - 1; buff[i] == ' '; i--) buff[i] = '\0';

	    add_message("%s\n", buff);

	    i = 0;

	}

    (void)sprintf(stat_file,"%s/%s",path2,names[j]);

    (void)stat(stat_file, &st);
    f_size=st.st_size/512;

     (void)sprintf(buff + i * max, "%5d %-*s",f_size || !st.st_size?f_size:1, max, names[j]);
     i++;
    }

  if (i) add_message("%s\n", buff);

  free((char *)names);
  free(name_buffer);
}

int tail(path)
    register char *path;
{
    char buff[1000];
    register FILE *f;
    struct stat st;
    int offset;

    path = check_file_name(path, 0,0);

    if (!path) return 0;

    f = fopen(path, "r");

    if(!f)
     if(errno == EMFILE) fatal("Max files open.\n");
	else return 0;

    if (fstat(fileno(f), &st) == -1)
	fatal("Could not stat an open file.\n");

    offset = st.st_size - 54 * 20;

    if (offset < 0) offset = 0;

    if (fseek(f,(long)offset, 0) == -1) fatal("Could not seek.\n");

    /* Throw away the first incomplete line. */

    if (offset > 0) (void)fgets(buff, sizeof buff, f);

    while(fgets(buff, sizeof buff, f))  add_message("%s", buff);

    (void)fclose(f);
    return 1;
}

VALUE *print_file(path, start, len,ret_methode)
    register char *path;
    unsigned int start, len;
    int ret_methode;
{
    char buff[257];
    register char *ret_string=0;
    register VALUE *ret;
    register FILE *f;
    int i,rd;

    if(!ret_methode && !command_giver) return &const0;

    if (len < 0 || start <0) return &const0;
  
    if(!len) len=100;
        
    path = check_file_name(path, 0,0);

    if (!path) return &const0;

    f = fopen(path, "r");
    if(!f)
      if(errno == EMFILE) fatal("Max files open.\n");
	else return &const0;

    if (!start) start = 1;

    for (i=1; len?i < start + len:1; i++) 
      {

	if(i>start+100 || !fgets(buff, sizeof buff-2, f)) break;

        if(i<start) continue;

           if(!ret_methode) add_message("%s", buff);
	      else
	 	if(!ret_string) ret_string=string_copy(buff);
		  else
		    {
	 	      ret_string=(char *)realloc(ret_string,strlen(buff)+strlen(ret_string)+1);
	 	      (void)strcat(ret_string,buff);
		    }
       }
    (void)fclose(f);
    if (i <= start) return &const0;
    if(i>100+start) add_message("***TRUNCATED***\n");
    if(!ret_methode) 
	return make_number(i-start,0.0,0);
      else 
       {
          ret=make_string(ret_string);
          free(ret_string);
          return ret;
        }
}

void remove_file(path)
    register char *path;
{
    path = check_file_name(path, 1,0);
    if (!path || unlink(path) == -1) add_message("No such file: %s\n", path);
}

/* warning, for sake of ease in the calls 'realm' is free'd here
   so don't 'free' it someplace else.
*/

void log_file(file, str,realm)
    register char *file, *str,*realm;
{
    FILE *f;
    char file_name[100];

    if (strchr(file, '/') || file[0] == '.' || strlen(file) > 30)
	error(ob_status[0]->ob->realm,"Illegal file name to log_file(%s)\n", file);

    if(realm) (void)sprintf(file_name,"%s/log/%s",realm,file);
     else (void)sprintf(file_name, "log/%s", file);

    f = fopen(file_name, "a");
    if(!f)
       if(errno == EMFILE) fatal("Max files open.\n");
	  else return;

    (void)fwrite(str,(unsigned int)strlen(str), 1, f);
    (void)fclose(f);
}
 
struct input_to *alloc_input_to() 
{
    register INPUT_TO *p;

    p=(INPUT_TO *)xalloc(sizeof (INPUT_TO));
#if 1
    (void)memset((char *)p,'\0',sizeof (INPUT_TO));
#else
    p->ob=0;
    p->fun_p=0;
    p->arg.type=T_INT;
    p->arg.u.string=0;
    p->arg.u.n_int=0;
    p->arg.u.n_float=0.0;
    p->arg.u.vec=0;
    p->arg.u.ob=0;
#endif
    return p;
}

void free_input_to(p)
    register INPUT_TO *p;
{

    if(p->ob) remove_ob_ref(p->ob);
    if(p->fun_p) free_fun_p(p->fun_p);

    if(p->arg.type==T_STRING) free_string(p->arg.u.string,"free_input_to: u.string");
      else if(p->arg.type==T_OBJECT) remove_ob_ref(p->arg.u.ob);
         else if(p->arg.type==T_POINTER) free_vector(p->arg.u.vec);

    free((char *)p);
}


struct sentence *sent_free = 0;
int tot_alloc_sentence;

struct sentence *alloc_sentence() 
{
    register SENTENCE *p;

    if (!sent_free)
      {
	int i;

	sent_free =(SENTENCE *)xalloc(cf.alloc_sent * sizeof (SENTENCE));

	for (i=0; i<cf.alloc_sent - 1; i++)
	    sent_free[i].next  = &sent_free[i+1];

	sent_free[cf.alloc_sent-1].next = 0;
    }

    p = sent_free;
    sent_free = sent_free->next;

    tot_alloc_sentence++;

#if 1
    (void)memset((char *)p,'\0',sizeof (SENTENCE));
#else
    p->ob=0;
    p->verb=0;
    p->next=0;
    p->type=0;
#endif

    return p;
}

void free_sentence(p)
    register SENTENCE *p;
{
    if(p->verb) free_string(p->verb,"free_sentence: verb");

    free_fun_p(p->fun_p);

    p->next=sent_free;
    sent_free=p;

    tot_alloc_sentence--;
}


int try_verb(s,data)
  register char *data;
  register SENTENCE *s;
{
	register VALUE *ret,*arg;
	register struct object_status *save_ob_status=0;
	jmp_buf erc;
    struct jmp_globals jmp;

	/*
	 * Now we have found a special sentence !
	 */
	last_verb=s;
	/*
	 * If the function is static and not defined by current object,
	 * then it will fail. If this is called directly from player input,
	 * then we set current_object so that static functions are allowed.
	 */
	if (!ob_status[0]->ob) 
       {
		  save_ob_status=ob_status[0];
		  ob_status[0]=alloc_ob_status();
		  ob_status[0]->ob=s->ob;
       }
	if(data) arg=make_string(data);
		else arg=make_number(0,0.0,0);

	if (command_giver->interactive) s->ob->wl->score++;

	set_jump_state(erc,&jmp);
	if(setjmp(error_recovery_context))
	 {
		clear_jump_state(&jmp);
		ret= &const0;
	 }
	else ret = apply(s->fun_p,0, s->ob, arg,0);
	reset_jump_state(erc);

    if(save_ob_status)
      {
        free_ob_status(ob_status[0]);
        ob_status[0]=save_ob_status;
      }

    if(!last_verb) s=0;
         else last_verb=0;

    if( (ret->type == T_INT && !ret->u.n_int) ||
        (ret->type==T_FLOAT && !ret->u.n_float) )
	       if(!s) return -1;
	          else return 0;

        return 1;
}

/*
 * Find the sentence for a command from the player.
 * Return success status.
 */

int player_parser(buff)
    register char *buff;
{
    register char *verb,*data;
    register VALUE *arg;
    register SENTENCE *s2=0;
    int flag=0,len,length,ret=0;

    if(!buff || buff[0]=='\0') return 0;

    /* strip leading spaces. */
    while(buff[0]==' ') buff++;

    if(!buff) return 0;

    /* strip trailing spaces. */
    for (length=strlen(buff); length>=0 && buff[length-1]==' ';length--)
	 buff[length-1]='\0';

    if (buff[0] == '\0') return 0;

#if 1
    if(special_parse(buff)) return 1;
#else
    if(command_giver->interactive)
      {
	extern OBJECT *find_link PROT((OBJECT *));
	register OBJECT *log_ob=find_link(command_giver);

	if(log_ob)
	 {
		register VALUE *ret;
		jmp_buf erc;
		struct jmp_globals jmp;

		set_jump_state(erc,&jmp);
		if(!setjmp(error_recovery_context))
		  {
		    ret=apply(0,"query_sysop",log_ob,&const0,0);
		    if(!numchk(ret,T_INT) && ret->u.n_int)
		      if(special_parse(buff))
			{
			   reset_jump_state(erc);
			   return 1;
			}
		  }
		else clear_jump_state(&jmp);
		reset_jump_state(erc);
	 }
     }
#endif

    if(!command_giver->verbs)
	{
           clear_notify();
	   notify_no_command();
	   return 1;
	}

    if(!strchr(buff, ' '))
	{
	   length=strlen(buff);
	   verb=string_copy(buff);
	   data=0;
	}
      else
        {
	   length = strlen(buff)-strlen(strchr(buff,' '));
	   verb=xalloc(length+1);
	   (void)strncpy(verb,buff,length);
  	   verb[length]=0;
	   data=string_copy(buff+length+1);
	} 
    clear_notify();

  get_verb_again:

    for(ret=0,s2=find_v_hash(command_giver,verb,flag);s2;s2=s2->next)
      {
	    if(flag==1 && s2->type) continue;
	    if(flag!=1)
              { 
	        if(s2->type==1)
		  { if(s2->verb[0]!=0 && strncmp(verb,s2->verb,strlen(s2->verb)))
			 break;
		  }
		else if(strcmp(verb,s2->verb)) break; 
	      }
	      else 
		  if(strncmp(verb,s2->verb,length)) break;
	    if(s2->type==1)
	      {
		 if(strlen(buff)==strlen(s2->verb))
			ret=try_verb(s2,0);
		  else ret=try_verb(s2,buff+strlen(s2->verb));		 
	      }
              else ret=try_verb(s2,data);

	    if(ret==1) break;

	    if(!command_giver)
	     {
	      ret=0;
	      flag=0;
	      break;
	     }

	    if(ret== -1) goto get_verb_again;

	    if(s2->verb[0]==0 && s2->next && s2->next->verb[0]!=0)
		{
		  flag= -1;
		  goto get_verb_again;
		}
      }

   if(flag!=1 && ret!=1 && command_giver && 
      !(command_giver->flags & VERB_MATCHING)) 
     {
	flag=1;
	goto get_verb_again;
     }

   free(verb);
   if(data) free(data);
   if(ret!=1)
     {
	notify_no_command();
	return 0;
     }
    return 1;
}

/*
 * Associate a command with function in this object.
 * The second argument is the command name.
 *
 * The optinal third argument is a flag that will state that the verb should
 * only match against leading characters.
 *
 * The object must be near the command giver, so that we ensure that the
 * sentence is removed when the command giver leaves.
 */
void add_action(str, cmd, flag)
    register char *str, *cmd;
    int flag;
{
    register SENTENCE *p;
    struct fun_p *fun_p;

    if (!command_giver || !str || !cmd) return;
    if (ob_status[0]->ob != command_giver &&
	ob_status[0]->ob->ref_ob[0] != command_giver &&
	ob_status[0]->ob->ref_ob[0] != command_giver->ref_ob[0] &&
	ob_status[0]->ob != command_giver->ref_ob[0])
      error(ob_status[0]->ob->realm,"add_action from object that was not present.\n");
    if (!strcmp(str, "exit"))
	error(ob_status[0]->ob->realm,"Illegal to define a command to the exit() function.\n");
	if(!strcmp(str,"init"))
	error(ob_status[0]->ob->realm,"Illegal to define a command to the init() function.\n");

    fun_p=find_fun_p(str,ob_status[0]->ob,0);

    if(!fun_p)
	   error(ob_status[0]->ob->realm,"Tried to add_action to a function that does not exist '%s'.\n",str);

    p = alloc_sentence();
    p->fun_p=fun_p;
    p->ob = ob_status[0]->ob;
    p->type = cmd[0]==0?(unsigned char)1:(unsigned char)flag;
    p->verb = make_shared_string(cmd);

    if(!command_giver->verbs)
     {
	int i,siz=sizeof (struct sentence *) * cf.vtable;

	v_table_overhead+=siz;
	v_table_count++;
	command_giver->verbs=(struct sentence **)xalloc(siz);
	for(i=0;i<cf.vtable;i++)
	   command_giver->verbs[i]=0;
     }

    add_v_hash(command_giver,p);
}

int remove_sent(ob, player,verb)
    register OBJECT *ob, *player;
    register char *verb;
{
    register SENTENCE *s,*next,*prev;
    int ret=0,index;

   if(!player->verbs) return 0;

    if(verb)
         return del_v_hash(player,verb);

   for(index=0;index<cf.vtable;index++)
    for(prev=0,s=player->verbs[index];s;s=next)
	  {
		 next=s->next;
		 if(!ob || (ob==s->ob))
		  {
			ret=1;
			if(prev) prev->next=next;
			  else player->verbs[index]=next;
			if(s==last_verb) last_verb=0;
			free_sentence(s);
			continue;
		  }
		prev=s;
	  }
	if(!ob) 
          {
	    free((char *)player->verbs);
	    player->verbs=0;
	    v_table_overhead-=sizeof (struct sentence *) *cf.vtable;
	    v_table_count--;
	  }
	return ret;
}

int special_parse(buff)
    register char *buff;
{
    if (strcmp(buff, "malloc") == 0) {
#ifdef MALLOC_umalloc
	malloc_status();
#else
# ifdef MALLOC_smalloc
	dump_malloc_data();
# else
#  ifdef MALLOC_gmalloc
  {
    extern unsigned int malloc_sbrk_used,malloc_sbrk_unused;
    extern int malloc_mem_free(),malloc_mem_used();

    add_message("Using Gnu malloc.\n");
    add_message("Sbrk used     : %8d\n",malloc_sbrk_used);
    add_message("Sbrk free     : %8d\n",malloc_sbrk_unused);
    add_message("Malloced Used : %8d (%8.2fk)\n",malloc_mem_used(),(float) malloc_mem_used()/1024);
    add_message("Malloced free : %8d (%8.2fk)\n",malloc_mem_free(),(float) malloc_mem_free()/1024);
    add_message("Total malloced: %8d (%8.2fk)\n",malloc_mem_free()+malloc_mem_used(),(float) (malloc_mem_free()+malloc_mem_used())/1024);
  }
#  endif
# endif
#endif
    return 1;
   }
 if (!strcmp(buff, "dumpallobj")) {
	extern void dumpstat();

        dumpstat();
	return 1;
    }
    if(!strncmp(buff, "proghash",8)) {
	extern show_prog_hash();
	char *pp;

	pp=strchr(buff,' ');
	show_prog_hash(pp?pp+1:0);
	return 1;
    }
    if(!strcmp(buff,"doreset")) {
	extern void move_current_reset();

	move_current_reset(1);
	return 1;
    }
   if(!strcmp(buff,"resetlist")) {
        extern void show_reset_list();

	show_reset_list();
 	return 1;
    }
    if (!strcmp(buff, "status")) 
      {
	VALUE *val;

	extern int num_swapped,num_resets_called;
	extern int total_bytes_swapped, total_num_prog_blocks;
	extern int tot_alloc_object,total_prog_block_size;
	extern int tot_alloced_inherit,tot_alloc_ob_bytes;
	extern int num_ob_in_free_list,tot_alloc_svalue;

	extern void heart_beat_status();
	extern void reset_status();
	extern void sys_stat();
	extern void show_o_table_status();
	extern void show_s_table_status();
	extern void show_p_table_status();

	add_message("\n--------------------- UriMud v%s System Statistics -------------------\n\n",VERSION);

	show_p_table_status();

	add_message("Objects currently loaded                    : %7d (%9d) bytes\n",
	tot_alloc_object,tot_alloc_ob_bytes);

	show_o_table_status();

	add_message("Inheritences                                : %7d (%9d) bytes\n",
	tot_alloced_inherit,tot_alloced_inherit * sizeof (struct inherit_list));

	add_message("Objects in free list                        : %7d (%9d) bytes\n",
	num_ob_in_free_list,num_ob_in_free_list*sizeof (struct object));

	add_message("Defined Verbs                               : %7d (%9d) bytes\n",
	 tot_alloc_sentence,tot_alloc_sentence * sizeof (struct sentence));

	add_message("Verb tables                                 : %7d (%9d) bytes\n",v_table_count,v_table_overhead);

	add_message("Global Varaibles                            : %7d (%9d) bytes\n",
	tot_alloc_svalue,tot_alloc_svalue*sizeof (struct svalue));

	add_message("Allocated Arrays                            : %7d (%9d) bytes\n",array_count,array_ovhead);

#if 0
	add_message("Programs swapped                            : %7d (%9d) bytes\n",
	num_swapped, total_bytes_swapped);
#endif

	show_s_table_status();
	(void)print_call_out_usage();
	heart_beat_status();
	reset_status();
	sys_stat();
	return 1;
      }
   if(!strcmp(buff,"lnodes"))
      {
	print_lnode_status();
	return 1;
      }
   if(!strcmp(buff,"hblist"))
     {
	struct link_list *hb;
	int i;
	extern struct link_list *hb_list;

	for(i=1,hb=hb_list;hb;hb=hb->next,i++)
	   add_message("%3d. %s\n",i,hb->ob->name);
	return 1;
     }
  if(!strcmp(buff,"callouts"))
    {
        extern void show_call_out_list();

	show_call_out_list();
	return 1;
    }
 return 0;
}

void print_local_commands(ob,ob2)
 register OBJECT *ob,*ob2;
 {
    register struct sentence *s;
    char buff[81];
    int index,c;

    if(!ob) ob=command_giver;
    buff[0]=0;

    if(!ob->verbs)
      {
        add_message("No verbs defined.\n");
        return;
      }

    add_message("Current local commands:\n");
  for(c=0,index=0;index<cf.vtable;index++)
    for (s = ob->verbs[index]; s; s = s->next)
	if(!ob2 || ob2==s->ob)
	 {
	   c++; 
	   if(strlen(buff)+strlen(s->verb)+2>78)
		  {
		    add_message("%s\n",buff);
		    buff[0]=0;
		  }
	   if(buff[0]) (void)strcat(buff,", ");

	   (void)strcat(buff,s->verb);
	 }
   if(buff[0]) add_message("%s\n",buff);
   add_message("\n%d verbs defined.\n",c);
}

FILE *pre_compile(c_name)
    register char *c_name;
{
    char *tmp;
    char cmd[1000];
    register REALMS *rel;
    FILE *f;

    extern unsigned sleep();

    rel=extract_realm(c_name);

    tmp=extract_wizard(c_name);

    if(tmp)
      {
	(void)sprintf(cmd,
        "%s -DVERSION=\"%s\" -I%s/%s/w/%s/include -I%s/%s/include -I%s/include %s",cf.cpp,
	 VERSION,cf.mud_lib,rel->realm_dir,tmp,cf.mud_lib,rel->realm_dir,cf.mud_lib,
	 c_name);
	free(tmp);
       }

     else if(rel!=sys_realm)
	(void)sprintf(cmd,
       	"%s -DVERSION=\"%s\" -I%s/%s/include -I%s/include %s",cf.cpp,VERSION, cf.mud_lib, rel->realm_dir,cf.mud_lib,c_name);

	else (void)sprintf(cmd,"%s -DVERSION=\"%s\" -I%s/include %s",cf.cpp,VERSION, cf.mud_lib,c_name);

     cmd[999]=0;  /* just in case we've built a string way too long. */

    if((f=popen(cmd,"r"))==NULL)
     {
       if(errno == EMFILE) fatal("load_object: Max files open.\n");
       fatal("load_object: Could not execute pre-compiler.\n");
     }

	sleep(1);
              /* We do this for two reasons:
		  1.  Give cpp a chance to run, this keeps compile time
                      up for loading small objects.

                  2.  To reduce CPU usage.  Removing this will more than
                      double the CPU usage while compiling.

                  Note: Very little perfomance is gained by not
                        having sleep(1) called at this time.
               */

    return f;
}

/*
 * Check that it is an legal path. No '..' are allowed.
 */
int legal_path(path)
    register char *path;
{
    register char *p;

    if (path == NULL || strchr(path, ' ')) return 0;
    if (path[0] == '/') path++;
    for(p = strchr(path, '.'); p; p = strchr(p+1, '.')) {
	if (p[1] == '.')
	    return 0;
    }
    return 1;
}

/*
 * Check that a file name is valid for read or write.
 * Also change the name as if the current directory was at the players
 * own directory.
 * This is done by functions in the mcp.
 */

char *check_file_name(file, writeflg,perms)
    int perms;
    char *file;
    int writeflg;
{
    register VALUE *ret;
    VALUE v,v2;
    register OBJECT *save_command_giver=command_giver;
    register struct object_status *save_ob_status[2];
    extern struct realms *load_realm;

    if(!file || file=="\0") return 0;

    v.type=T_STRING;
    v.u.string=file;

    v2.type=T_STRING;
    if(load_realm) v2.u.string=load_realm->realm_dir;
	else if(ob_status[0]->ob) v2.u.string=ob_status[0]->ob->realm->realm_dir;
	   else if(command_giver) v2.u.string=command_giver->realm->realm_dir;
		else v2.u.string=sys_realm->realm_dir;

    /*
     * We don't have to free the string in ret. This is done
     * by the garbage collection.
     */

	if(perms)
         {
	   command_giver=0;
	   save_ob_status[0]=ob_status[0];
	   save_ob_status[1]=ob_status[1];
	   ob_status[1]=0;
	   ob_status[0]=alloc_ob_status();
	 }

        if (writeflg) ret = apply(mcp.fun_p[1],0, mcp.ob, &v,&v2);
    	   else ret = apply(mcp.fun_p[0],0, mcp.ob, &v,&v2);

	if(perms)
         {
		command_giver=save_command_giver;
		free_ob_status(ob_status[0]);
		if(ob_status[1]) free_ob_status(ob_status[1]);
    		ob_status[0]=save_ob_status[0];
		ob_status[1]=save_ob_status[1];
	 }

    if (strchk(ret,1) || !legal_path(ret->u.string)) return 0;

    return ret->u.string;
}

char game_is_being_shut_down;

void shutdowngame(sig) 
 int sig;
{
	extern int kill(),getpid();

          (void)signal(SIGHUP , SIG_DFL);
          (void)signal(SIGTERM, SIG_DFL);
          (void)signal(SIGKILL, SIG_DFL);
          (void)signal(SIGILL , SIG_DFL);
          (void)signal(SIGIOT , SIG_DFL);
#ifndef sun
          (void)signal(SIGSEGV, SIG_DFL);
#endif
#ifndef AIX
          (void)signal(SIGABRT, SIG_DFL);
#endif

    shut_down_game=1;
    if(sig) debug_message(sys_realm,"shutdown on signal: %d\n",sig);
    shout_string("Game driver shouts: UriMUD shutting down immediately.\n",1);
    save_wiz_file();
    remove_all_players();
    ipc_remove();
    if(!sig) exit(0);
      else kill(getpid(),sig);
}

/*
 * Call this one when there is only little memory left. It will start
 * Armageddon.
 */
void slow_shut_down(minutes)
    int minutes;
{
    struct object *ob;
    struct value v;
	OBJECT *save_current_object=ob_status[0]->ob;
	jmp_buf erc;
	struct jmp_globals jmp;

    ob_status[0]->ob=0;
    shut_down_game=1;
    ob = find_object("closed/shut");
    if (!ob) {
	shout_string("Game driver shouts: Out of memory.\n",1);
	ob_status[0]->ob = save_current_object;
	shutdowngame(0);
	return;
    }
    shout_string("Game driver shouts: The memory is getting low !\n",1);
    v.type = T_INT;
    v.u.n_int = minutes;
	set_jump_state(erc,&jmp);
	if(setjmp(error_recovery_context))
	  {
		 clear_jump_state(&jmp);
		 shout_string("Game driver shouts: Could not execute delayed shutdown!\n",1);
		 shutdowngame(0);
	  }
      else (void)apply(0,"shut", ob, &v,0);
	reset_jump_state(erc);
    ob_status[0]->ob=save_current_object;
}

int match_string(match, str)
    register char *match, *str;
{
    int i;

 again:
    if (*str == '\0' && *match == '\0') return 1;
    switch(*match) 
	  {
    	case '?':
		 	if (*str == '\0') return 0;
			str++;
			match++;
			goto again;
    	case '*':
			match++;
			if (*match == '\0') return 1;
			for (i=0; str[i] != '\0'; i++)
	    		if (match_string(match, str+i)) return 1;
			return 0;
    	case '\0':
			return 0;
    	case '\\':
			match++;
			if (*match == '\0') return 0;
			/* Fall through ! */
    	default:
			if (*match == *str) 
			  {
	    		match++;
	    		str++;
	    		goto again;
			  }
			return 0;
    	}
}

struct value *slice_array(p,from,to)
     register VALUE *p;
     int from;
     int to;
{
  register VECTOR *d;
  register VALUE *v;
  int cnt;
  
  if (from>=p->u.vec->size) return 0;  /* Slice starts above array */
  if (to>=p->u.vec->size || to== -1) to=p->u.vec->size-1;
  if (to<from) return 0; 
  
  v =(struct value *)allocate_array((unsigned short)(to-from+1));
  d= v->u.vec;
  for (cnt=from;cnt<=to;cnt++) 
{
      assign_svalue(&d->item[cnt-from],0,&p->u.vec->item[cnt]);
}

  return v;
}

/* Concatenation of two arrays into one
*/

VALUE *add_array(v,v2)
     register VALUE *v;
     register VALUE *v2;
{
  register VALUE *new;
  int c,i;
  
  if (pointchk(v) || pointchk(v2)) return &const0;

  if(v->u.vec->size+v2->u.vec->size<1) return v;

  new=allocate_array((unsigned short)(v->u.vec->size+v2->u.vec->size));

  i=0;

  if(v->u.vec->size)
     for(;i<v->u.vec->size;i++)
	assign_svalue(&new->u.vec->item[i],0,&v->u.vec->item[i]);

  if(v2->u.vec->size)
     for(c=0;c<v2->u.vec->size;c++,i++)
	assign_svalue(&new->u.vec->item[i],0,&v2->u.vec->item[c]);

  return new;
}


int search_array(vec,val,index)
   register VECTOR *vec;
   register VALUE *val;
   int index;
{

  int i;

  for(i=index;i<vec->size;i++)
    {
       if(vec->item[i].type==T_OBJECT && !check_object(vec->item[i].u.ob))
          assign_svalue(&vec->item[i],&const0,0);

       if(vec->item[i].type!=val->type) continue;

       switch(val->type)
	 {
	    case T_OBJECT:
		if(vec->item[i].u.ob==val->u.ob) return i;
		break;
	    case T_INT:
		if(vec->item[i].u.n_int==val->u.n_int) return i;
		break;
	    case T_FLOAT:
		if(vec->item[i].u.n_float==val->u.n_float) return i;
		break;
	    case T_STRING:
		if(!strcmp(vec->item[i].u.string,val->u.string)) return i;
		break;
	    case T_POINTER:
		if(vec->item[i].u.vec==val->u.vec) return i;
		break;
	    default:
		error(sys_realm,"Bad data type to search_array()\n");
	}
   }
 return -1;
}

void del_element(p,res)
     register VECTOR *p;
     int res;
{
  int next;


  char *arr,*arr2;
  int i;

  i=sizeof (struct svalue);

  arr=(char *)p->item;
  arr2=(char *)&p->item[res+1];

  memcpy(arr+(i*res),arr2,(i*(p->size-res)));
  memset(arr+(i*(p->size-1)),'\0',i);

  p->size--;
}
 
static char *find_percent(str)
    register char *str;
{
    while(1) {
	str = strchr(str, '%');
	if (!str) return 0;
	if (str[1] != '%') return str;
	str++;
    }
}

VALUE *inter_sscanf(l)
    LNODE *l;
{
    register VALUE *p;
    register char *fmt,*cp;
    char *in_string, *match;
    int number_of_matches;

    /*
     * First get the string to be parsed.
     */
    p = get_one_expr(&l, "sscanf");
    if (strchk(p,0) || !p->u.string) do_error("1st","sscanf");
    in_string =p->u.string;
    if (!in_string) return &const0;
    /*
     * Now get the format description.
     */
    p = get_one_expr(&l, "sscanf");
    if (strchk(p,1)) do_error("2nd","sscanf");
    fmt = p->u.string;
    /*
     * First, skip and match leading text.
     */
    for (cp=find_percent(fmt); fmt != cp; fmt++, in_string++) 
	if (in_string[0] == '\0' || fmt[0] != in_string[0]) return &const0;
    /*
     * For every % or substring in the format.
     */
    for (number_of_matches=0; l && l->a1; l = l->a2) {
	int i, type;

	if (fmt[0] == '\0') {
	    /*
	     * We have reached end of the format string.
	     * If there are any chars left in the in_string,
	     * then we put them in the last variable (if any).
	     */
	    if (in_string[0]) {
		assign(l->a1, make_string(in_string));
		number_of_matches++;
	    }
	    break;
	}
	if (fmt[0] != '%') fatal("Should be a %% now !\n");
	switch(fmt[1])
	  {
			case 's':
				type=1;
				break;
			case 'd':
				type=2;
				break;
			case 'c':
				type=3;
				break;
			case 'f':
				type=4;
				break;
			default:
			    error(ob_status[0]->ob->realm,"Bad type : '%%%c' in sscanf fmt string.\n",fmt[1]);
		  }
	fmt += 2;
    if(type!=1)
	 {
	  switch(type)
		{
			case 2:
			  {
    			char *tmp = in_string;
    			int tmp_num;
			extern int strtol();

    			tmp_num = strtol(in_string, &in_string, 0);
    			if(tmp == in_string) break;
				number_of_matches++;
    			assign(l->a1, make_number(tmp_num,0.0,0));
				break;
			  }
			case 3:
			  {
				unsigned char tmp_num;
				
				tmp_num=in_string[0];
				in_string++;
				number_of_matches++;
				assign(l->a1,make_number(tmp_num,0.0,0));
				break;
			  }
			case 4:
			  {
    			double tmp_num;

				tmp_num=atof(in_string);
				if(!tmp_num && in_string[0]!='0') break;
				while(strchr("0123456789.+-Ee",in_string[0])) in_string++;
				number_of_matches++;
    			assign(l->a1, make_number(0,tmp_num,1));
				break;
			  }
			default:
			  error(ob_status[0]->ob->realm,"bad numeric type set in sccanf()\n");
		  }
   		while(fmt[0] && fmt[0] == in_string[0]) fmt++, in_string++;
	    if (fmt[0] != '%') break;
	    continue;
	   }
	/*
	 * Now we have the string case.
	 */
	cp = find_percent(fmt);

	if (cp == fmt)
	    error(ob_status[0]->ob->realm,"Illegal to have 2 adjacent %'s in fmt string in sscanf.");
	if (!cp) cp = fmt + strlen(fmt);

	/*
	 * First case: There was no extra characters to match.
	 * Then this is the last match.
	 */
	if (cp == fmt) {
	    assign(l->a1, make_string(in_string));
	    number_of_matches++;
	    break;
	}
	for (i=0,match=0; in_string[i]; i++) 
	    if (!strncmp(in_string+i, fmt, cp - fmt)) {
		/*
		 * Found a match !
		 */
		match = xalloc(i+1);
		(void)strncpy(match, in_string, i);
		in_string += i + cp - fmt;
		match[i] = '\0';
		assign(l->a1, make_string(match));
		number_of_matches++;
		free(match);
		fmt = cp;	/* Advance fmt to next % */
		break;
	    }

	if (fmt == cp) continue;	/* If match, then do continue. */
	/*
	 * No match was found. Then we stop here, and return
	 * the result so far !
	 */
	break;
    }
    return make_number(number_of_matches,0.0,0);
}

OBJECT *find_player(player,flag)
	register char *player;
	int flag;
{
	int i;
	extern int num_player;

	for(i=1;i<=num_player;i++)
	 {
		register OBJECT *ob;

		if(!flag)
		 if(!check_realm_access(0,get_interactive_object(i,flag)))
		   continue;

		ob=get_interactive_object(i,flag);
		if(ob->living_name && !strcmp(player,ob->living_name))
		     return ob;
	  }
   return 0;
}
		
struct value *users(flag)
  int flag;
{
    register OBJECT *ob;
    extern int num_player; /* set by comm1.c */
    int i;
    unsigned short num;
    register VALUE *ret;

    for (i = 0,num=0; i < num_player; i++) 
      {
	ob=get_interactive_object(i+1,flag);
	if(!flag && ob->realm!=ob_status[0]->ob->realm) continue;
	num++;
      }
    
    if(!num) return &const0;

    ret = allocate_array(num);

    for (num=0, i = 0; i < num_player; i++) 
      {
	SVALUE v;

	ob=get_interactive_object(i+1,flag);
	if(!flag && ob->realm!=ob_status[0]->ob->realm) continue;

	v.type=T_OBJECT;
	v.u.ob=ob;
	assign_svalue(&ret->u.vec->item[num++],0,&v);
	ob->num_ref++;
      }

    return ret;
}

struct value *explode_string(str, del)
    register char *str,*del;
{
    register char *buff;
    char *q;
    int num,len;
    register VALUE *ret;
    SVALUE v;

    len=strlen(del);

    /* Treat each delimiter as a field, even if it contains no data. */
   
    /*
     * Find number of occurences of the delimiter 'del'.
     */

	for(num=1,q=strstr(str,del);q;num++,q+=len,q=strstr(q,del));

	if(num>cf.max_array_size)
	    {
		 free(del);
		 error(ob_status[0]->ob->realm,"explode: Tried to make an array larger than the sytsem max: %d.\n",cf.max_array_size);
	   }

    ret = allocate_array(num);

    num=0;

    while(1)
	 {
	     q=strstr(str,del);

	     if(!q)
	      {
                v.type=T_STRING;
	        v.u.string=str;
		assign_svalue(&ret->u.vec->item[num],0,&v);
		break;
	      }
	    else
	      {
		int len2;

	        len2=strlen(str)-strlen(q);
		buff=xalloc(len2+1);

		(void)strncpy(buff,str,len2);
		buff[len2]=0;

		v.type=T_STRING;
		v.u.string=buff;
		assign_svalue(&ret->u.vec->item[num++],0,&v);

		free(buff);
	        str= &q[len];
	      }
         }

    free(del);
    return ret;
}

struct value *implode_string(arr, del)
    register VECTOR *arr;
    char *del;
{
    int i, num,len;
    register char *p;
    register VALUE *ret;

    len=strlen(del);

    for(i=0,num=1;i<arr->size;i++)
	if(arr->item[i].type == T_STRING)
		num+=strlen(arr->item[i].u.string);

    p = xalloc((len*arr->size) + num);

    for (i=0, num=0; i < arr->size; i++) 
       {
	 if (arr->item[i].type == T_STRING)
	   {
              if(num)
                 {
                    (void)strcat(p,del);
		    (void)strcat(p, arr->item[i].u.string);
		    continue;
		 }

    	       (void)strcpy(p,arr->item[i].u.string);
	       num=1;
	       continue;
	   }

         if(num)
	   {
	      (void)strcat(p,del);
	      continue;
	   } 

         (void)strcpy(p,del);
	 num=1;
       }

    ret=make_string(p);
    free(del);
    free(p);
    return ret;
}

void write_file(file, str,perms)
    register char *file;
    register char *str;
    int perms;
{
    register FILE *f;

    if(!perms) file = check_file_name(file, 1,0);
    if(file)
     {
       f = fopen(file, "a");
       if(f)
        {
          fwrite(str, strlen(str), 1, f);
          fclose(f);
	}
       else 
          if(errno == EMFILE) fatal("Max files open.\n");
     }
}

int file_size(file)
    register char *file;
{
    struct stat st;

    if (!file || stat(file, &st) == -1) return -1;
    if (S_IFDIR & st.st_mode) return -2;
    return st.st_size;
}

VALUE *inter_sprintf(l)
    LNODE *l;
{
    extern char *save_array PROT((VECTOR *,int));

    register VALUE *p;
    register LNODE *l2;
    register char *cp,*fmt,*get_fmt,*buff=0;
    int num_arg=0;
    register LNODE *ret;

    if(!l) error(ob_status[0]->ob->realm,"sprintf: No data\n");
	ret=l->a1;
    /*
     * Now get the format description.
     */
	p = get_one_expr(&l, "sprintf");
	if(!p) do_error("1st","sprintf");
    p = get_one_expr(&l, "sprintf");
    if (strchk(p,0)) do_error("2nd","sprintf");
	if(strchk(p,1))
	 {
		assign(ret,p);
		return p;
	 }
    fmt=get_fmt=string_copy(p->u.string);
	while(fmt)
	 {
		char err_count=0;

    	p=get_one_expr(&l,0);
    	if(!p || !strchr(fmt,'%'))
	  {
		if(!buff) 
 		  {
		    buff=xalloc(strlen(fmt)+1);
		    buff[0]=0;
		  }
		   else buff=(char *)realloc(buff,strlen(buff)+strlen(fmt)+1);

	  	(void)strcat(buff+strlen(buff),fmt);
		break;
	  }
	 else
	  {
		register char *cp=fmt,*cp2;
		char found=0;
		char buf2[23000];

		while(cp && found!=2)
	         {   
			cp=strchr(cp,'%');
			if(!cp || strlen(cp)<2) break;
			if(cp[1]=='%') {cp+=2;continue;}
			  else  {if(found<2) cp++;found++;}
		      }
		    if(!found)
		      {
			if(!buff) buff=xalloc(strlen(fmt)+1);
			  else buff=(char *)realloc(buff,strlen(buff)+strlen(fmt)+1);
			(void)sprintf(buff+strlen(buff),fmt);
			break;
		      }
		     else
		      {
	     	  	if(cp)
		  	{
				int len;

		  		len=strlen(fmt)-strlen(cp);
				if(!len) len=strlen(fmt);
		  		cp2=xalloc(len);
		  		(void)strncpy(cp2,fmt,len-1);
				cp2[len-1]=0;
				fmt+=len-1;
				switch(p->type)
				  {
				 case T_STRING:
					  (void)sprintf(buf2,cp2,p->u.string);
					break;
				 case T_OBJECT:
					   (void)sprintf(buf2,cp2,p->u.ob->name);
						break;
				 case T_POINTER:
				    {
					register char *arr_buff=save_array(p->u.vec,1);

					(void)sprintf(buf2,cp2,arr_buff);
					free(arr_buff);
				     }
					break;
				 case T_INT:
					  (void)sprintf(buf2,cp2,p->u.n_int);
					break;
				 case T_FLOAT:
					  (void)sprintf(buf2,cp2,p->u.n_float);
			  }
	  		free(cp2);
			if(err_count) 
		       {
			  if(buff) free(buff);
			  free(get_fmt);
			  error(ob_status[0]->ob->realm,"Bad conversion type on data type.\n");
			 }
		 }
		else
		 {
		   cp2=string_copy(fmt);
			switch(p->type)
			  {
				 case T_STRING:
			  		   (void)sprintf(buf2,cp2,p->u.string);
					break;
				 case T_OBJECT:
					  (void)sprintf(buf2,cp2,p->u.ob->name);
					break;
				 case T_POINTER:
					{
					  register char *arr_buff=save_array(p->u.vec,1);

					  (void)sprintf(buf2,cp2,arr_buff);
					  free(arr_buff);
					}
				    break;
				 case T_INT:
					  (void)sprintf(buf2,cp2,p->u.n_int);
					break;
				 case T_FLOAT:
					  (void)sprintf(buf2,cp2,p->u.n_float);
				  }
				free(cp2);
				fmt=0;
				if(err_count) 
				{
				    if(buff) free(buff);
				    free(get_fmt);
				    error(ob_status[0]->ob->realm,"Bad conversion to data type.\n");
				}
			    }
			  if(!buf2) continue;
			  if(!buff)
			   {
				 buff=xalloc(strlen(buf2)+1);
				 (void)strcpy(buff,buf2);
			   }
			  else 
			  {
				 buff=(char *)realloc(buff,strlen(buff)+strlen(buf2)+1);
				 (void)strcat(buff,buf2);
			  }
			 continue;
	       }
		 }
       }
	if(!buff) p= &const0;
      else
	    {
		  p=make_string(buff);
		  free(buff);
		}
    assign(ret,p);
    free(get_fmt);
	return p;
}

/*	Find first occurence of str2 in str1 and return a pointer to it */
#ifdef STRSTR
char *strstr(str1, str2)
register char *str1, *str2;
{
	register char *Sptr, *Tptr;
	int len = strlen(str1) -strlen(str2) + 1;

	if (*str2)
		for (; len > 0; len--, str1++) {
			if (*str1 != *str2)
				continue;

			for (Sptr = str1, Tptr = str2; *Tptr != '\0'; Sptr++, Tptr++)
				if (*Sptr != *Tptr)
					break;

			if (*Tptr == '\0')
				return str1;
		}

	return 0;
}
#endif

void intr_grep(grep_string)
	register char *grep_string;
{
	register char *cmd,*grep_cmd,*grep_file;
	char grep_output[200];
	register VALUE *grep_value,*name;
	register VECTOR *grep_arr;
	int file_size,i,base;
	FILE *p;

	if(!check_object(command_giver)) return;

	name=apply(0,"query_user_name",command_giver,&const0,0);
	if(strchk(name,1)) error(ob_status[0]->ob->realm,"Must be a user to use grep.\n");

	grep_value=explode_string(grep_string,string_copy(" "));

	if(pointchk(grep_value)) error(ob_status[0]->ob->realm,"Bad grep format.\n");

	grep_arr=grep_value->u.vec;

	if(grep_arr->size<2) error(ob_status[0]->ob->realm,"wrong number of arguments to grep().\n");

	if(grep_arr->item[0].u.string[0]!='-') base=2;
		else base=3;

	for(file_size=strlen(grep_arr->item[base-1].u.string)+1,i=base-1;i<grep_arr->size;i++)
	 {
		VALUE *new_path,v;

		if(grep_arr->item[i].type!=T_STRING)
			error(ob_status[0]->ob->realm,"Bad (%d) argument to grep, must be a string.\n",i);
		v.type=T_STRING;
		v.u.string=grep_arr->item[i].u.string;

		new_path=apply(mcp.fun_p[2],0,mcp.ob,&v,0);

		if(strchk(new_path,1))
			 error(ob_status[0]->ob->realm,"Bad file path '%s' to grep().\n",v.u.string);

		if(!check_file_name(new_path->u.string,0,0))
			error(ob_status[0]->ob->realm,"grep: No read permision on:%s.\n",grep_arr->item[i].u.string);
		file_size+=strlen(new_path->u.string);
		assign_svalue(&grep_arr->item[i],new_path,0);
	 }
	grep_file=xalloc(file_size+1);

	(void)strcpy(grep_file,grep_arr->item[base-1].u.string);

	for(i=base;i<grep_arr->size;i++)
	 {
		(void)strcat(grep_file," ");
		(void)strcat(grep_file,grep_arr->item[i].u.string);
	 }

	if(base==3)
 	 {
   	    grep_cmd=xalloc(strlen(grep_arr->item[0].u.string)+strlen(grep_arr->item[1].u.string)+5);
	    (void)sprintf(grep_cmd,"%s '%s' ",grep_arr->item[0].u.string,grep_arr->item[1].u.string);
	 }
	else
 	 {
   	    grep_cmd=xalloc(strlen(grep_arr->item[0].u.string)+5);
	    (void)sprintf(grep_cmd,"'%s' ",grep_arr->item[0].u.string);
	 }

	(void)sprintf(grep_output,"%s/log/%s.grep",command_giver->realm->realm_dir, name->u.string);
	cmd=xalloc(strlen(grep_file)+strlen(grep_cmd)+strlen(cf.grep)+strlen(grep_output)+8);
	(void)sprintf(cmd,"%s %s %s >>%s &",cf.grep,grep_cmd,grep_file+1,grep_output);
	free(grep_cmd);
	free(grep_file);

        if((p=popen(cmd,"r"))==NULL)
         {
           if(errno == EMFILE) fatal("Max files open.\n");
           error(ob_status[0]->ob->realm,"grep: could not execute.\n");
         }

	(void)pclose(p);
	free(cmd);
}

int intr_system(file,cmd_string)
	register char *cmd_string,*file;
{
	FILE *f;
        int res;
	register char *cmd;

	if(file[0]!='/') error(sys_realm,"Bad path to system().\n");
        if(!check_file_name(file,0,0)) 
		error(sys_realm,"system: No read permsion\n");

	cmd=xalloc(strlen(file)+(cmd_string?strlen(cmd_string)+24:24));
	(void)strcpy(cmd,file+1);

	if(cmd_string)
	  {
		(void)strcat(cmd," ");
		(void)strcat(cmd,cmd_string);
	  }

	(void)strcat(cmd," > log/System.sys &");

        if((f=popen(cmd,"r"))==NULL)
         {
	   free(cmd);
           if(errno == EMFILE) fatal("Max files open.\n");
	   error(sys_realm,"system: Unable to execute command.\n");
         }

        (void)pclose(f);
	return 1;
}
